from unittest.mock import call, patch

from django.contrib.auth.models import AnonymousUser
from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rq.exceptions import NoSuchJobError
from SolrClient import SolrResponse
from SolrClient.exceptions import SolrError

from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestSearchApi(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.corpus.indexable = True
        cls.corpus.types.filter(slug="page").update(indexable=True)
        cls.corpus.save()

    def build_solr_response(
        self,
        docs=[],
        query="",
        facets={
            "element_type": [],
            "element_worker": [],
            "transcription_worker": [],
            "classification_name": [],
            "classification_worker": [],
            "metadata_name": [],
            "metadata_type": [],
            "metadata_worker": [],
            "entity_type": [],
            "entity_worker": []
        },
        nb_results=None
    ):
        return SolrResponse({
            "responseHeader": {
                "QTime": 1,
                "params": {
                    "q": query
                }
            },
            "response": {
                "numFound": nb_results if nb_results else len(docs),
                "start": 0,
                "numFoundExact": True,
                "docs": docs
            },
            "facet_counts": {
                "facet_fields": facets,
            }
        })

    @override_settings(ARKINDEX_FEATURES={"search": False})
    def test_search_unavailable(self):
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ["Search features are not available on this instance."])

    @override_settings(ARKINDEX_FEATURES={"search": True})
    def test_corpus_not_found(self):
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "No Corpus matches the given query."})

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_corpus_no_permission(self, has_access_mock):
        private_corpus = Corpus.objects.create(name="private", public=False)
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": private_corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(AnonymousUser(), private_corpus, Role.Guest.value, skip_public=False))

    @override_settings(ARKINDEX_FEATURES={"search": True})
    def test_corpus_not_indexable(self):
        self.corpus.indexable = False
        self.corpus.save()
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), [f"Corpus {self.corpus.id} is not indexable."])

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_index_not_found(self, mock_solr):
        mock_solr.collections.exists.return_value = False
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": f"Corpus index project-{self.corpus.id} not found."})

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_empty_query(self, mock_solr):
        mock_solr.collections.exists.return_value = True
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"query": ["This field is required."]})

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_wrong_query(self, mock_solr):
        # Mock SolrClient
        mock_solr.collections.exists.return_value = True
        mock_solr.query.side_effect = SolrError()

        payload = {
            "sources[]": ["element"],
            "query": "*)"
        }
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}), payload)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Query "*)" is not valid.'])

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search(self, mock_solr):
        self.maxDiff = None
        collection_name = f"project-{self.corpus.id}"
        docs = [{
            "id": "document_id",
            "parent_id": "parent_id",
            "parent_name": "1",
            "parent_type": "Page",
            "element_id": "element_id",
            "element_text": "3",
            "element_type": "Paragraph",
            "element_worker": "1234567890_A worker",
            "element_image": "http://image.url",
            "transcription_id": "transcription_id",
            "transcription_text": "A text",
            "transcription_confidence": 0.2,
            "transcription_worker": "1234567890_A worker",
            "classification_id": "classification_id",
            "classification_name": "my class",
            "classification_text": "my class",
            "classification_confidence": 0.1,
            "classification_worker": "1234567890_A worker",
            "metadata_id": "metadata_id",
            "metadata_name": "A metadata",
            "metadata_text": "My value",
            "metadata_type": "text",
            "metadata_worker": "1234567890_A worker",
            "entity_id": "entity_id",
            "entity_text": "An entity",
            "entity_type": "person",
            "entity_worker": "1234567890_A worker",
        }]

        # Mock SolrClient
        mock_solr.collections.exists.return_value = True
        query = '(element_text:("I search" OR "Found") OR transcription_text:("I search" OR "Found") OR classification_text:("I search" OR "Found")) AND (metadata_name:"folio" AND entity_type:"person")'
        solr_response = self.build_solr_response(docs=docs, query=query)
        mock_solr.query.return_value = solr_response

        payload = {
            "sources[]": ["element", "transcription", "classification"],
            "metadata_name": "folio",
            "entity_type": "person",
            "query": '"I search" OR "Found"',
        }
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}), payload)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "next": None,
            "previous": None,
            "results": docs,
            "facets": solr_response.get_facets()
        })

        # Check Solr call
        (index_name, ), _ = mock_solr.collections.exists.call_args
        self.assertEqual(index_name, collection_name)
        (index_name, args), kwargs = mock_solr.query.call_args
        self.assertEqual(index_name, collection_name)
        # The order in which the sources appear in the query is random, so we just check that all the sources are there
        q = args.pop("q")
        self.assertIn('element_text:("I search" OR "Found")', q)
        self.assertIn('transcription_text:("I search" OR "Found")', q)
        self.assertIn('classification_text:("I search" OR "Found")', q)
        self.assertDictEqual(args, {
            "start": 0,
            "rows": 20,
            "sort": "score desc",
            "facet": True,
            "facet.field": [
                "element_type",
                "element_worker",
                "transcription_worker",
                "classification_name",
                "classification_worker",
                "metadata_name",
                "metadata_type",
                "metadata_worker",
                "entity_type",
                "entity_worker"
            ]
        })

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_only_facets(self, mock_solr):
        docs = [{
            "id": "doc_id",
            "parent_id": "parent_id",
            "parent_name": "1",
            "parent_type": "Page",
            "element_id": "element_id",
            "element_text": "3",
            "element_type": "Paragraph"
        }]
        facets = {
            "element_type": [
                "Paragraph", 1,
                "Page", 0,
                "Text line", 0
            ],
            "element_worker": [],
            "transcription_worker": [],
            "classification_name": [],
            "classification_worker": [],
            "metadata_name": [],
            "metadata_type": [],
            "metadata_worker": [],
            "entity_type": [],
            "entity_worker": []
        }

        # Mock SolrClient
        mock_solr.collections.exists.return_value = True
        solr_response = self.build_solr_response(docs=docs, facets=facets)
        mock_solr.query.return_value = solr_response

        payload = {
            "query": "Test",
            "only_facets": True
        }
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}), payload)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 0,
            "number": 1,
            "next": None,
            "previous": None,
            "results": None,
            "facets": solr_response.get_facets()
        })

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_pagination(self, mock_solr):
        # Mock SolrClient
        mock_solr.collections.exists.return_value = True
        solr_response = self.build_solr_response(nb_results=42)
        mock_solr.query.return_value = solr_response

        payload = {
            "query": "Test",
            "page": 2
        }
        response = self.client.get(reverse("api:corpus-search", kwargs={"pk": self.corpus.id}), payload)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 42,
            "number": 2,
            "next": f"http://testserver/api/v1/corpus/{self.corpus.id}/search/?page=3&query=Test",
            "previous": f"http://testserver/api/v1/corpus/{self.corpus.id}/search/?page=1&query=Test",
            "results": [],
            "facets": solr_response.get_facets()
        })

        # Check Solr call
        (_, args), _ = mock_solr.query.call_args
        self.assertEqual(args["start"], 20)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_sort_element_name(self, mock_solr):
        mock_solr.collections.exists.return_value = True
        solr_response = self.build_solr_response()
        mock_solr.query.return_value = solr_response

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:corpus-search", kwargs={"pk": self.corpus.id}),
                {"query": "test", "sort": "element_name"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 0,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [],
            "facets": solr_response.get_facets()
        })

        (_, args), _ = mock_solr.query.call_args
        self.assertEqual(args["sort"], "element_text asc")

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.api.search.solr")
    def test_search_sort_unknown(self, mock_solr):
        mock_solr.collections.exists.return_value = True

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:corpus-search", kwargs={"pk": self.corpus.id}),
                {"query": "test", "sort": "element_atomic_number"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "sort": ['"element_atomic_number" is not a valid choice.']
        })
        self.assertFalse(mock_solr.query.called)

    def test_build_index_requires_auth(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_build_index_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_build_index_requires_corpus_admin(self, has_access_mock):
        self.corpus.memberships.update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Admin.value, skip_public=False))

    @override_settings(ARKINDEX_FEATURES={"search": False})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    def test_build_index_requires_search_feature(self, reindex_trigger_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["Building search index is unavailable due to the search feature being disabled."]
        })
        self.assertFalse(reindex_trigger_mock.called)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index_existing_running_job(self, job_mock, reindex_trigger_mock):
        existing_job = job_mock()
        existing_job.ended_at = None
        job_mock.fetch.return_value = existing_job
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": [f"A job is already running to build search index on corpus {self.corpus.id}."]
        })
        self.assertFalse(reindex_trigger_mock.called)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index_non_indexable_corpus(self, job_mock, reindex_trigger_mock):
        self.corpus.indexable = False
        self.corpus.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["This project is not indexable."]
        })
        self.assertFalse(reindex_trigger_mock.called)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index_non_indexable_element_types(self, job_mock, reindex_trigger_mock):
        self.corpus.types.update(indexable=False)
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["There are no indexable element types for this project."]
        })
        self.assertFalse(reindex_trigger_mock.called)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index_existing_finished_job(self, job_mock, reindex_trigger_mock):
        """In case a job exists but is finished, remove it and index the corpus again"""
        existing_job = job_mock()
        existing_job.ended_at = "2000-01-01"
        job_mock.fetch.return_value = existing_job
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:build-search-index", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"drop": True})
        self.assertListEqual(reindex_trigger_mock.call_args_list, [call(
            corpus_id=self.corpus.id,
            drop=True,
            user_id=self.user.id,
            description=f"Reindexing project {self.corpus.id}",
            job_id=f"reindex-{self.corpus.id}",
        )])
        # The previous Job has been deleted
        self.assertTrue(existing_job.delete.called)

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index(self, job_mock, reindex_trigger_mock):
        job_mock.fetch.side_effect = NoSuchJobError
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:build-search-index", kwargs={"pk": self.corpus.id}),
                data={"drop": True},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"drop": True})
        self.assertListEqual(reindex_trigger_mock.call_args_list, [call(
            corpus_id=self.corpus.id,
            drop=True,
            user_id=self.user.id,
            description=f"Reindexing project {self.corpus.id}",
            job_id=f"reindex-{self.corpus.id}",
        )])

    @override_settings(ARKINDEX_FEATURES={"search": True})
    @patch("arkindex.documents.tasks.reindex_corpus.delay")
    @patch("arkindex.project.rq_overrides.Job")
    def test_build_index_drop_false(self, job_mock, reindex_trigger_mock):
        job_mock.fetch.side_effect = NoSuchJobError
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:build-search-index", kwargs={"pk": self.corpus.id}),
                data={"drop": False},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"drop": False})
        self.assertListEqual(reindex_trigger_mock.call_args_list, [call(
            corpus_id=self.corpus.id,
            drop=False,
            user_id=self.user.id,
            description=f"Reindexing project {self.corpus.id}",
            job_id=f"reindex-{self.corpus.id}",
        )])
