import copy
from datetime import datetime, timezone
from io import BytesIO
from itertools import combinations
from unittest.mock import patch, seal

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import FINAL_STATES, Agent, AgentMode, Farm, State
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState
from arkindex.users.models import Role


@override_settings(PONOS_LOG_TAIL=42)
class TestTaskUpdate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            cls.process.run()
        cls.task1, cls.task2, cls.task3 = cls.process.tasks.all().order_by("depth")

        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")

        cls.farm = Farm.objects.first()
        cls.docker_agent = Agent.objects.create(
            mode=AgentMode.Docker,
            farm=cls.farm,
            last_ping=datetime.now(timezone.utc),
            cpu_cores=42,
            cpu_frequency=42e8,
            ram_total=42e3
        )
        cls.slurm_agent = Agent.objects.create(
            mode=AgentMode.Slurm,
            farm=cls.farm,
            last_ping=datetime.now(timezone.utc),
        )

    @property
    def docker_task_transitions(self):
        return (
            (State.Unscheduled, State.Pending),
            (State.Pending, State.Error),
            (State.Pending, State.Running),
            (State.Running, State.Completed),
            (State.Running, State.Failed),
            (State.Running, State.Error),
            (State.Running, State.Cancelled),
            (State.Stopping, State.Stopped),
            (State.Stopping, State.Error),
        )

    @property
    def slurm_task_transitions(self):
        return self.docker_task_transitions + (
            (State.Pending, State.Completed),
            (State.Pending, State.Failed),
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level")
    def test_requires_process_admin_corpus(self, get_max_level_mock):
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                get_max_level_mock.return_value = role.value if role else None

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.ponos.models.base64.encodebytes")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_worker_activities_update(self, get_user_mock, s3_mock, token_mock):
        """
        When a task is updated to a final state, only the remaining worker activities that are not in
        the processed (or already in error) state are updated to the error state.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        token_mock.side_effect = [b"12345", b"78945",]

        test_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
            activity_state=ActivityState.Ready
        )
        init_run = test_process.worker_runs.create(version=WorkerVersion.objects.get(worker__slug="initialisation"), ttl=0)
        test_run = test_process.worker_runs.create(version=self.recognizer, parents=[init_run.id], ttl=0)

        test_process.run()

        elem_1, elem_2, elem_3, elem_4 = self.corpus.elements.all()[0:4]
        activity_processed = WorkerActivity.objects.create(
            element_id=elem_1.id,
            state=WorkerActivityState.Processed,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        activity_queued = WorkerActivity.objects.create(
            element_id=elem_2.id,
            state=WorkerActivityState.Queued,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        activity_started = WorkerActivity.objects.create(
            element_id=elem_3.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        activity_error = WorkerActivity.objects.create(
            element_id=elem_4.id,
            state=WorkerActivityState.Error,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )

        test_task = test_process.tasks.get(worker_run=test_run)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        # Get the allowed transitions to a final state
        cases = [item for item in self.docker_task_transitions if item[1] in FINAL_STATES]

        for state_from, state_to in cases:
            with self.subTest(state_from=state_from, state_to=state_to):

                test_task.state = state_from
                test_task.save()

                resp = self.client.put(
                    reverse("api:task-details", args=[test_task.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                test_task.refresh_from_db()
                activity_started.refresh_from_db()
                activity_error.refresh_from_db()
                activity_queued.refresh_from_db()
                activity_processed.refresh_from_db()

                self.assertEqual(test_task.state, state_to)
                self.assertEqual(activity_processed.state, WorkerActivityState.Processed)
                self.assertEqual(activity_error.state, WorkerActivityState.Error)
                self.assertEqual(activity_queued.state, WorkerActivityState.Error)
                self.assertEqual(activity_started.state, WorkerActivityState.Error)

    @patch("arkindex.ponos.models.base64.encodebytes")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_worker_activities_update_worker_run(self, get_user_mock, s3_mock, token_mock):
        """
        When a task is updated to a final state, only the worker activities corresponding to that task's worker
        run are updated.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        token_mock.side_effect = [b"12345", b"78945",]

        test_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
            activity_state=ActivityState.Ready
        )
        init_version = WorkerVersion.objects.get(worker__slug="initialisation")
        init_run = test_process.worker_runs.create(version=init_version, ttl=0)
        test_run = test_process.worker_runs.create(version=self.recognizer, parents=[init_run.id], ttl=0)

        test_process.run()

        elem = self.corpus.elements.first()
        init_activity = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=init_version,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        reco_activity = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )

        test_task = test_process.tasks.get(worker_run=test_run)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        test_task.state = State.Running
        test_task.save()

        resp = self.client.put(
            reverse("api:task-details", args=[test_task.id]),
            data={"state": State.Completed.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        test_task.refresh_from_db()
        init_activity.refresh_from_db()
        reco_activity.refresh_from_db()

        self.assertEqual(test_task.state, State.Completed)
        self.assertEqual(init_activity.state, WorkerActivityState.Started)
        self.assertEqual(reco_activity.state, WorkerActivityState.Error)

    @patch("arkindex.ponos.models.base64.encodebytes")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_worker_activities_update_chunks(self, get_user_mock, s3_mock, token_mock):
        """
        When a task is updated to a final state, the worker activities corresponding to that task's worker run
        are only updated if there are no other tasks for that same worker run (different chunks) that are not in
        a final state.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        token_mock.side_effect = [b"12345", b"78945", b"77975"]

        test_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
            chunks=2,
            activity_state=ActivityState.Ready
        )
        test_run = test_process.worker_runs.create(version=self.recognizer, ttl=0)

        test_process.run()

        elem = self.corpus.elements.first()
        reco_activity = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )

        test_task_1, test_task_2 = test_process.tasks.filter(worker_run=test_run).all()

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        test_task_2.state = State.Running
        test_task_2.save()

        resp = self.client.put(
            reverse("api:task-details", args=[test_task_2.id]),
            data={"state": State.Completed.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        test_task_2.refresh_from_db()
        reco_activity.refresh_from_db()
        test_task_1.refresh_from_db()

        # Only one of the tasks is in a final state: the WorkerActivityState did not change
        self.assertEqual(test_task_2.state, State.Completed)
        self.assertEqual(test_task_1.state, State.Unscheduled)
        self.assertEqual(reco_activity.state, WorkerActivityState.Started)

        for final_state in FINAL_STATES:
            with self.subTest(final_state=final_state):
                # Put test_task_1 in a final state as well
                test_task_1.state = final_state
                test_task_2.state = State.Running
                test_task_1.save()
                test_task_2.save()

                resp = self.client.put(
                    reverse("api:task-details", args=[test_task_2.id]),
                    data={"state": State.Completed.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                test_task_2.refresh_from_db()
                reco_activity.refresh_from_db()
                test_task_1.refresh_from_db()

                # The remaining worker activity has been updated to the Error state
                self.assertEqual(test_task_2.state, State.Completed)
                self.assertEqual(test_task_1.state, final_state)
                self.assertEqual(reco_activity.state, WorkerActivityState.Error)

    @override_settings(PUBLIC_HOSTNAME="https://arkindex.localhost")
    @patch("arkindex.ponos.models.base64.encodebytes")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_worker_activities_update_same_worker_version(self, get_user_mock, s3_mock, token_mock):
        """
        When a task is updated to a final state, only the remaining worker activities corresponding to that task's worker
        run are updated to the error state, not activities belonging to other workers runs with the same worker version.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        token_mock.side_effect = [b"12345", b"78945", b"77975", b"12995"]

        test_model = Model.objects.create(name="Generic model", public=False)
        test_model_version = test_model.versions.create(
            state=ModelVersionState.Available,
            tag="Test",
            hash="A" * 32,
            archive_hash="42",
            size=1337,
        )
        test_configuration = self.recognizer.worker.configurations.create(
            name="Recognizer configuration", configuration={"value": "test"}
        )

        test_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
            activity_state=ActivityState.Ready
        )
        test_run_1 = test_process.worker_runs.create(
            version=self.recognizer,
            ttl=0,
        )
        test_run = test_process.worker_runs.create(
            version=self.recognizer,
            model_version_id=test_model_version.id,
            ttl=0,
        )
        test_run_2 = test_process.worker_runs.create(
            version=self.recognizer,
            model_version_id=test_model_version.id,
            configuration_id=test_configuration.id,
            ttl=0,
        )

        test_process.run()

        elem = self.corpus.elements.first()
        target_activity = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            model_version=test_model_version,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        activity_1 = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )
        activity_2 = WorkerActivity.objects.create(
            element_id=elem.id,
            state=WorkerActivityState.Started,
            worker_version=self.recognizer,
            model_version=test_model_version,
            configuration=test_configuration,
            process_id=test_process.id,
            started=datetime.now(timezone.utc)
        )

        test_task = test_process.tasks.get(worker_run=test_run)
        task_1 = test_process.tasks.get(worker_run=test_run_1)
        task_2 = test_process.tasks.get(worker_run=test_run_2)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        test_task.state = State.Running
        task_1.state = State.Stopped
        task_2.state = State.Failed
        test_task.save()
        task_1.save()
        task_2.save()

        resp = self.client.put(
            reverse("api:task-details", args=[test_task.id]),
            data={"state": State.Completed.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        test_task.refresh_from_db()
        target_activity.refresh_from_db()
        activity_1.refresh_from_db()
        activity_2.refresh_from_db()

        self.assertEqual(test_task.state, State.Completed)
        self.assertEqual(target_activity.state, WorkerActivityState.Error)
        self.assertEqual(activity_1.state, WorkerActivityState.Started)
        self.assertEqual(activity_2.state, WorkerActivityState.Started)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_from_docker_agent_allowed_transitions(self, notify_mock, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in self.docker_task_transitions:
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.put(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_to)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_from_slurm_agent_allowed_transitions(self, notify_mock, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.slurm_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in self.slurm_task_transitions:
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.put(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_to)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_from_docker_agent_forbidden_transitions(self, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Mock authenticating from an agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in combinations(State, 2):
            if (state_from, state_to) in self.docker_task_transitions:
                continue
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.put(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(resp.json(), {
                    "state": [f"Transition from state {state_from} to state {state_to} is forbidden."]
                })
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_from)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_from_slurm_agent_forbidden_transitions(self, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Mock authenticating from an agent
        custom_user = copy.copy(self.slurm_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in combinations(State, 2):
            if (state_from, state_to) in self.slurm_task_transitions:
                continue
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.put(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(resp.json(), {
                    "state": [f"Transition from state {state_from} to state {state_to} is forbidden."]
                })
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_from)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_running_sets_started(self, notify_mock, s3_mock, get_user_mock):
        """
        Updating a task's state to Running sets its `started` date
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        self.assertIsNone(self.task1.started)
        self.assertIsNone(self.task1.finished)
        self.task1.state = State.Pending
        self.task1.save()

        with self.assertNumQueries(4):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Running.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Running)
        self.assertIsNotNone(self.task1.started)
        self.assertIsNone(self.task1.finished)

        data = resp.json()
        self.assertIsNotNone(data["started"])
        self.assertIsNone(data["finished"])


    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_sets_finished(self, notify_mock, s3_mock, get_user_mock):
        """
        Updating a task's state to a final state sets its `finished` date
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for state in FINAL_STATES:
            with self.subTest(state=state):
                self.task1.started = datetime.now(timezone.utc)
                self.task1.finished = None
                self.task1.state = State.Stopping.value if state == State.Stopped else State.Running.value
                self.task1.save()

                with self.assertNumQueries(8 if state == State.Completed else 7):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": state.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)
                self.assertIsNotNone(self.task1.started)
                self.assertIsNotNone(self.task1.finished)

                data = resp.json()
                self.assertIsNotNone(data["started"])
                self.assertIsNotNone(data["finished"])


    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_final_state_finished_requires_started(self, notify_mock, s3_mock, get_user_mock):
        """
        Updating a task's state to a final state does not set its `finished` date when `started` is not set
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for state in FINAL_STATES:
            with self.subTest(state=state):
                self.task1.started = None
                self.task1.finished = None
                self.task1.state = State.Stopping.value if state == State.Stopped else State.Running.value
                self.task1.save()

                with self.assertNumQueries(8 if state == State.Completed else 7):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": state.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)
                self.assertIsNone(self.task1.started)
                self.assertIsNone(self.task1.finished)

                data = resp.json()
                self.assertIsNone(data["started"])
                self.assertIsNone(data["finished"])
