from django.contrib import admin
from django.db.models import Max
from enumfields.admin import EnumFieldListFilter

from arkindex.process.models import (
    DataFile,
    Process,
    Worker,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
)
from arkindex.project.admin import ArchivedListFilter


class DataFileInline(admin.StackedInline):
    model = Process.files.through
    extra = 0


class WorkerRunInline(admin.StackedInline):
    model = Process.versions.through
    raw_id_fields = ("version", "configuration", "model_version")
    readonly_fields = ("summary", )
    extra = 0


class ElementInline(admin.StackedInline):
    model = Process.elements.through
    raw_id_fields = ("element", )
    readonly_fields = ("element", )
    extra = 0


class ProcessAdmin(admin.ModelAdmin):
    list_display = ("id", "creator", "corpus", "state", "mode")
    list_filter = [("mode", EnumFieldListFilter), ]
    fieldsets = (
        (None, {"fields": ("id", "name", "creator", "corpus", "state", "mode", "activity_state", "template")}),
        ("Elements filters", {
            "fields": ("element", "element_type", "folder_type", "name_contains")
        }),
    )
    readonly_fields = ("id", "name", "corpus", "state", "activity_state", "template")
    raw_id_fields = ("element", )
    inlines = [DataFileInline, WorkerRunInline, ElementInline]

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs) \
            .prefetch_related("tasks", "corpus") \
            .annotate(last_run=Max("tasks__run"))


class DataFileAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "corpus")
    list_filter = ("corpus", )
    fields = ("id", "name", "size", "content_type", "corpus")
    readonly_fields = ("id", "size", )
    inlines = [DataFileInline, ]


class WorkerInline(admin.StackedInline):
    model = Worker


class WorkerVersionInline(admin.StackedInline):
    show_change_link = True
    model = WorkerVersion
    fields = ("state", "configuration", "docker_image_iid")
    readonly_fields = ("state", "configuration", "docker_image_iid")

    def get_queryset(self, *args, **kwargs):
        """
        Prefetch workers as they are used in versions inline labels.
        Using a proxy model to override `__str__` is possible but removes the Change link
        as the model does not appear as registered.
        """
        return super().get_queryset(*args, **kwargs).prefetch_related("worker")


class WorkerConfigurationInline(admin.StackedInline):
    model = WorkerConfiguration
    readonly_fields = ("configuration_hash", )


class WorkerAdmin(admin.ModelAdmin):
    list_display = ("name", "slug", "archived", "type")
    list_filter = (ArchivedListFilter, )
    fields = ("id", "name", "slug", "type", "description", "public", "archived")
    readonly_fields = ("id", )
    inlines = [WorkerVersionInline, WorkerConfigurationInline]

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).select_related("type")


class WorkerTypeAdmin(admin.ModelAdmin):
    list_display = ("id", "slug", "display_name", "created")
    fields = ("id", "slug", "display_name")
    readonly_fields = ("id", )


class WorkerVersionAdmin(admin.ModelAdmin):
    list_display = ("id", "worker", "version", "feature")
    list_filter = ("worker", ("feature", EnumFieldListFilter), )
    fields = (
        "id",
        "worker",
        "version",
        "revision_url",
        "tag",
        "branch",
        "configuration",
        "model_usage",
        "gpu_usage",
        "docker_image_iid",
        "state",
        "feature",
    )
    readonly_fields = ("id", )


class WorkerConfigurationAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "worker")
    list_filter = ("worker", )
    readonly_fields = ("id", "configuration_hash")


admin.site.register(Process, ProcessAdmin)
admin.site.register(DataFile, DataFileAdmin)
admin.site.register(Worker, WorkerAdmin)
admin.site.register(WorkerType, WorkerTypeAdmin)
admin.site.register(WorkerVersion, WorkerVersionAdmin)
admin.site.register(WorkerConfiguration, WorkerConfigurationAdmin)
