from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.db.models import Max
from teklia_toolbox.config import ConfigParser

from arkindex.process.models import ArkindexFeature, FeatureUsage, Worker, WorkerType, WorkerVersion, WorkerVersionState


def parse_config():
    parser = ConfigParser()
    parser.add_option("version", type=str)


    features_parser = parser.add_subparser("features")
    for feature in ArkindexFeature:
        feature_parser = features_parser.add_subparser(feature.value, allow_extra_keys=False, default={})
        feature_parser.add_option("image", type=str)
        feature_parser.add_option("command", type=str, default=None)

    return parser.parse(settings.BASE_DIR / "system_workers.yml")


class Command(BaseCommand):
    help = "Update the workers used to provide Arkindex features to the versions compatible with this release."

    def get_system_worker(self, feature: ArkindexFeature) -> Worker:
        """
        Update or create a `system` worker for this feature.
        Creates the `system` worker type if it does not exist.
        """
        try:
            conflicting_worker = Worker.objects.exclude(type__slug="system").filter(slug=feature.value).get()
            self.stdout.write(f"Worker {conflicting_worker.name} ({conflicting_worker.id}) uses the same slug with a different worker type.")
            self.stdout.write("Please update this worker's type to `system` or change its slug.")
            raise CommandError(f"Cannot create system worker with slug {feature.value} due to a conflict.")
        except Worker.DoesNotExist:
            pass

        worker_type, created = WorkerType.objects.get_or_create(
            slug="system",
            defaults={"display_name": "System"},
        )

        if created:
            self.stdout.write(f"Created new System worker type ({worker_type.id})")

        worker, created = Worker.objects.get_or_create(
            type=worker_type,
            slug=feature.value,
            defaults={
                "name": feature.name,
                "public": True,
            },
        )

        if created:
            self.stdout.write(f"Created new {worker.name} system worker")
        else:
            self.stdout.write(f"Using existing system worker {worker.name}")
            self.update_existing_worker(worker)

        return worker

    def update_existing_worker(self, worker: Worker) -> None:
        """
        Check that an existing worker has the correct attributes to have one of its versions provide a feature
        """
        updated = False

        if worker.archived is not None:
            self.stdout.write("Unarchiving worker")
            worker.archived = None
            updated = True

        if not worker.public:
            self.stdout.write("Marking worker as public")
            worker.public = True
            updated = True

        if updated:
            worker.save()
        else:
            self.stdout.write("Worker is up to date")

    def update_or_create_version(self, worker: Worker, feature: ArkindexFeature, docker_image: str, docker_command: str = None) -> None:
        """
        On a specified worker, assigns an existing version to a feature or creates a new one.
        Expects that no version is already assigned to this feature on any worker.
        """
        assert worker.archived is None, "Cannot assign a version on an archived worker to a feature"

        # Since there can be worker versions using the same Docker image but with different commands,
        # we need to filter by command as well as by Docker image.
        docker_command_filter = {"configuration__docker__command": docker_command} if docker_command else {"configuration__docker__command__isnull": True}
        # docker_image_iid is not unique, so we sort by state and take the most recent one.
        # Sorting by state means we prefer `available` versions first, then `created`, then `error`.
        worker_version = worker.versions.filter(
            docker_image_iid=docker_image,
            # We ignore existing versions with attributes that could interfere with the features,
            # like blocking a start or retry, or wasting resources.
            configuration__user_configuration__isnull=True,
            gpu_usage=FeatureUsage.Disabled,
            model_usage=FeatureUsage.Disabled,
            **docker_command_filter
        ).order_by("state", "-updated").first()

        if worker_version:
            self.stdout.write(f"Assigning existing worker version {worker_version.id} to the feature")
            worker_version.feature = feature

            if docker_command is not None:
                if "docker" not in worker_version.configuration:
                    worker_version.configuration["docker"] = {}
                worker_version.configuration["docker"]["command"] = docker_command
                self.stdout.write(f"Setup docker command: {docker_command}")

            if worker_version.state != WorkerVersionState.Available:
                self.stdout.write("Marking the worker version as available")
                worker_version.state = WorkerVersionState.Available

            worker_version.save()
            self.stdout.write(self.style.SUCCESS(f"Using existing worker version {worker_version.id}"))
            return

        self.stdout.write("Creating new worker version")
        max_version = worker.versions.aggregate(max_version=Max("version"))["max_version"] or 0
        worker_version = worker.versions.create(
            docker_image_iid=docker_image,
            feature=feature,
            state=WorkerVersionState.Available,
            version=max_version + 1,
            configuration={"docker": {"command": docker_command}} if docker_command else {}
        )
        self.stdout.write(self.style.SUCCESS(f"Using new worker version {worker_version.id}"))

    def check_existing_version(self, worker_version: WorkerVersion, feature: ArkindexFeature, docker_image: str, docker_command: str = None) -> None:
        self.stdout.write(f"Current worker version: {worker_version.id} ({worker_version.docker_image_iid})")
        valid = True

        if worker_version.docker_image_iid != docker_image:
            self.stdout.write(self.style.WARNING("This version has an invalid Docker image."))
            valid = False

        if worker_version.worker.archived is not None:
            self.stdout.write(self.style.WARNING("This version is part of an archived worker."))
            valid = False

        if worker_version.docker_command is None:
            if docker_command is not None:
                self.stdout.write(self.style.WARNING(f"This version requires the following Docker command: {docker_command}."))
                valid = False

        else:
            if docker_command is not None and docker_command == worker_version.docker_command:
                self.stdout.write("This version uses the required Docker command.")
            else:
                self.stdout.write(self.style.WARNING("This version uses a custom Docker command which could interfere with the feature."))
                valid = False

        if worker_version.required_user_configuration_fields:
            self.stdout.write(self.style.WARNING("This version requires a custom worker configuration which could interfere with the feature."))
            valid = False

        if valid:
            # Ensure the worker is public, and keep using this worker version
            self.update_existing_worker(worker_version.worker)
            self.stdout.write(self.style.SUCCESS(f"Worker version for {feature.name} is up to date"))
            return

        self.stdout.write("Unassigning feature from the current version")
        worker_version.feature = None
        worker_version.save(update_fields=["feature"])

        # Create a new version on the existing worker, even if it isn't a system worker
        worker = worker_version.worker
        if worker.archived is not None:
            # except if it is archived, since the new version would be invalid
            worker = self.get_system_worker(feature)
        else:
            self.update_existing_worker(worker_version.worker)

        self.update_or_create_version(worker, feature, docker_image, docker_command)

    @transaction.atomic
    def update_feature(self, feature: ArkindexFeature, config: dict):
        self.stdout.write(f" {feature.name} ".center(80, "─"))
        self.stdout.write(f"Using {config['image']} to provide {feature.name}")

        try:
            worker_version = WorkerVersion.objects.get_by_feature(feature)
            self.check_existing_version(worker_version, feature, config["image"], config["command"])
        except WorkerVersion.DoesNotExist:
            worker = self.get_system_worker(feature)
            self.update_or_create_version(worker, feature, config["image"], config["command"])

    def handle(self, *args, **options):
        config = parse_config()
        for feature_value, feature_config in config["features"].items():
            feature = ArkindexFeature(feature_value)
            self.update_feature(feature, feature_config)
